import { getTermios } from './termios.ts';

// ----------------------------------------------------------------------------
// Runtime adapter interfaces
// ----------------------------------------------------------------------------
export enum RunTimeType {
	Bun = 'bun',
	Deno = 'deno',
	Unknown = 'common',
}

/**
 * The native functions for getting/setting terminal settings
 */
export interface IFFI {
	/**
	 * Get the existing termios settings (for canonical mode)
	 */
	tcgetattr(fd: number, termiosPtr: unknown): number;

	/**
	 * Update the termios settings
	 */
	tcsetattr(fd: number, act: number, termiosPtr: unknown): number;

	/**
	 * Update the termios pointer with raw mode settings
	 */
	cfmakeraw(termiosPtr: unknown): void;

	/**
	 * Convert a TypedArray to an opaque pointer for ffi calls
	 */
	// deno-lint-ignore no-explicit-any
	getPointer(ta: any): unknown;
}

export interface ITerminalSize {
	rows: number;
	cols: number;
}

/**
 * Runtime-specific terminal functionality
 */
export interface ITerminal {
	/**
	 * The arguments passed to the program on launch
	 */
	argv: string[];
	/**
	 * The generator function returning chunks of input from the stdin stream
	 */
	inputLoop(): AsyncGenerator<Uint8Array, void, unknown>;

	/**
	 * Get the size of the terminal
	 */
	getTerminalSize(): Promise<ITerminalSize>;

	/**
	 * Pipe a string to stdout
	 */
	writeStdout(s: string): Promise<void>;
}

/**
 * Runtime-specific file handling
 */
export interface IFIO {
	openFile(path: string): Promise<string>;
	openFileSync(path: string): string;
	appendFile(path: string, contents: string): Promise<void>;
}

/**
 * The common interface for runtime adapters
 */
export interface IRuntime {
	/**
	 * The name of the runtime
	 */
	name: RunTimeType;

	/**
	 * Runtime-specific FFI
	 */
	ffi: IFFI;

	/**
	 * Runtime-specific terminal functionality
	 */
	term: ITerminal;

	/**
	 * Runtime-specific file system io
	 */
	file: IFIO;

	/**
	 * Set up an event handler
	 *
	 * @param eventName - The event to listen for
	 * @param handler - The event handler
	 */
	onEvent: (eventName: string, handler: (e: Event) => void) => void;

	/**
	 * Set a beforeExit/beforeUnload event handler for the runtime
	 * @param cb - The event handler
	 */
	onExit(cb: () => void): void;

	/**
	 * Stop execution
	 *
	 * @param code
	 */
	exit(code?: number): void;
}

// ----------------------------------------------------------------------------
// Misc runtime functions
// ----------------------------------------------------------------------------

let scrollRuntime: IRuntime | null = null;

/**
 * Kill program, displaying an error message
 * @param s
 */
export async function die(s: string | Error): Promise<void> {
	(await getTermios()).disableRawMode();
	console.error(s);
	(await getRuntime()).exit();
}

/**
 * Determine which Typescript runtime we are operating under
 */
export function getRuntimeType(): RunTimeType {
	let runtime = RunTimeType.Unknown;

	if ('Deno' in globalThis) {
		runtime = RunTimeType.Deno;
	}
	if ('Bun' in globalThis) {
		runtime = RunTimeType.Bun;
	}

	return runtime;
}

export async function getRuntime(): Promise<IRuntime> {
	if (scrollRuntime === null) {
		const runtime = getRuntimeType();
		const path = `../${runtime}/mod.ts`;

		const pkg = await import(path);
		if ('default' in pkg) {
			scrollRuntime = pkg.default;
		}
	}

	return Promise.resolve(scrollRuntime!);
}

/**
 * Import a runtime-specific module
 *
 * eg. to load "src/bun/mod.ts", if the runtime is bun,
 * you can use like so `await importForRuntime('index')`;
 *
 * @param path - the path within the runtime module
 */
export const importForRuntime = async (path: string) => {
	const runtime = getRuntimeType();
	const suffix = '.ts';
	const base = `../${runtime}/`;

	const pathParts = path.split('/')
		.filter((part) => part !== '' && part !== '.' && part !== suffix)
		.map((part) => part.replace(suffix, ''));

	const cleanedPath = pathParts.join('/');
	const importPath = base + cleanedPath + suffix;

	return await import(importPath);
};

/**
 * Import the default export for a runtime-specific module
 * (this is just a simple wrapper of `importForRuntime`)
 *
 * @param path - the path within the runtime module
 */
export const importDefaultForRuntime = async (path: string) => {
	const pkg = await importForRuntime(path);
	if ('default' in pkg) {
		return pkg.default;
	}

	return null;
};
